/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef GPIOPORTPIN_H
#define GPIOPORTPIN_H

#include <stdint.h>
#include "gpiopin.h"
#include "gpioport.h"

/**
 * GPIO pin that can be used as input or output and allows changing direction at runtime. The GpioPortPin works with
 * any GpioPort interface, as for example implemented by the PortA, PortB etc. classes that represent the AVR's
 * GPIO ports.
 * @note The GpioPortPin forwards its functions directly to the associated GpioPort device. That means for example when
 *       getting a value while configured as output, it will return the state that was set. When it is required to get
 *       the level that is actually applied, the direction must be changed to input first. Depending on the external
 *       circuit it also may take some time until a reliable value can be read.
 */
class GpioPortPin : public GpioPin
{
public:
	/**
	* Constructor.
	* @param port Pointer to GPIO port the pin is located on.
	* @param index Zero based index of the pin within the port
	*        (e.g. index 2 on PortB results in the pin commonly called PB2).
	*/
	GpioPortPin(GpioPort *port, uint8_t index) :
		m_port(port), m_mask(1u << index)
	{
	}

	virtual bool getValue(void) const override
	{
		return (m_port->getValue() & m_mask) != 0;
	}

	virtual void setValue(bool value) override
	{
		m_port->setValue(value ? m_mask : 0, m_mask);
	}

	virtual void setDirection(Direction dir) override
	{
		m_port->setDirection(dir, m_mask);
	}

	virtual void setPullup(bool state) override
	{
		m_port->setPullup(state ? m_mask : 0, m_mask);
	}

private:
	GpioPort * const m_port;
	const uint8_t m_mask;
};

#endif
