/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef INPUTPORTPIN_H
#define INPUTPORTPIN_H

#include <stdint.h>
#include "inputpin.h"
#include "inputport.h"

/**
 * Input pin that can be asked for its state, returning a boolean value. The InputPortPin works with any InputPort
 * interface, for example a GpioPort.
 * @note You have to make sure that pins being used for a InputPortPin to operate on must be configured as input.
 *       The InputPortPin doesn't know about setting a direction. So if you plan to use a GpioPort for this, first set
 *       it up correctly using its setDirection() method. This also includes activating or deactivating internal pullup
 *       resistors.
 */
class InputPortPin : public InputPin
{
public:
	/**
	* Constructor.
	* @param port Pointer to input port the pin is located on.
	* @param index Zero based index of the pin within the port
	*        (e.g. index 2 on PortB results in the pin commonly called PB2).
	*/
	InputPortPin(InputPort *port, uint8_t index) :
		m_port(port), m_mask(1u << index)
	{
	}

	virtual bool getValue(void) const override
	{
		return (m_port->getValue() & m_mask) != 0;
	}

private:
	InputPort * const m_port;
	const uint8_t m_mask;
};

#endif
