/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef ITEMCOLLECTION_H
#define ITEMCOLLECTION_H

/**
 * The ItemCollection template class allows to create a collection of arbitrary items. Using the forEach() method,
 * a custom callback method is invoked for each item in the collection. A context is available which allows the
 * callback's owner to pass arguments to the callback method which are specific to this forEach() invocation and
 * thus cannot be stored statically at the caller site.
 * @param BASETYPE Type of items to be collected.
 * @param COUNT Maximum number of items in the collection.
 * @note Due to the way the ItemCollection stores elements internally, the order of elements is hard to predict. An
 *       application designer should not make any assumptions about the order in which elements are handled using
 *       the forEach() method.
 * @note avr-gcc currently doesn't support std::function, and when using "normal" function pointers to catch
 *       lambda expressions you cannot use captures.
 * @todo This one needs thorough testing! Especially compact() and its dependencies.
 */
template <class BASETYPE, uint16_t COUNT> class ItemCollection
{
public:
	/**
	 * Type for callback method for forEach() method.
	 * @param item Current item in iteration.
	 * @param context Pointer to context that was passed to forEach() invocation.
	 */
	typedef void (*ForEachCallback)(BASETYPE &item, const void *context);

	/**
	 * Constructor.
	 */
	ItemCollection(void) : m_maxIndex(0)
	{
	}

	/**
	 * Adds an item to the collection. This method does not enforce the item to be unique.
	 * See addUnique() if this is a requirement.
	 * @param item Item to be added.
	 * @result True if item was added successfully, false otherwise.
	 */
	bool add(BASETYPE &item)
	{
		bool result = false;

		const uint16_t pos = nextEmptySlot(0);
		if (pos != COUNT)
		{
			insert(pos, item);
			result = true;
		}

		return result;
	}

	/**
	 * Adds an item to the collection. This method enforcse the item to be unique.
	 * See add() if duplicates should be allowed.
	 * @param item Item to be added.
	 * @result True if item was added successfully, false otherwise.
	 */
	bool addUnique(BASETYPE &item)
	{
		bool result = false;

		if (find(item) == COUNT)
		{
			result = add(item);
		}

		return result;
	}

	/**
	 * Removes an item from the collection. If there are duplicates, only one
	 * instance is removed.
	 * @param item Item to be removed.
	 * @note Removing items creates "holes" inside the internal data structure which affect the
	 *       performance of forEach() runs. These holes can be removed using the compact() method.
	 */
	void remove(BASETYPE &item)
	{
		const uint16_t pos = find(item);
		if (pos != COUNT)
		{
			m_entries[pos].used = false;
		}
	}

	/**
	 * Compacts the internal data storage by filling "holes" with items from other positions. This can be
	 * useful after removing items from the collection to improve the performance of forEach() runs.
	 * @warning This operation affects the order of elements and may disturb ongoing forEach() runs.
	 */
	void compact(void)
	{
		if (m_maxIndex > 0)
		{
			uint16_t readPos = prevFilledSlot(m_maxIndex - 1);
			uint16_t writePos = nextEmptySlot(0);

			while (writePos < readPos)
			{
				m_entries[writePos] = m_entries[readPos];
				m_entries[readPos].used = false;
				writePos = nextEmptySlot(writePos + 1);
				readPos = prevFilledSlot(readPos - 1);
			}

			m_maxIndex = readPos;
		}
	}

	/**
	 * Iterates through all items in the collection and invokes the specified callback method on
	 * each item. The context is passed to each callback invocation to allow the caller passing
	 * information which needs to be specific for this forEach() run.
	 * @param action Callback method to be called for each item.
	 * @param context Pointer to context to be passed to each callback invocation.
	 * @note Removing items from the collection in context of the callback method is safe, but
	 *       when adding items it is not guaranteed if they are part of this forEach() run or not.
	 */
	void forEach(ForEachCallback action, const void *context = nullptr)
	{
		for (uint16_t i = 0; i < m_maxIndex; i++)
		{
			if (m_entries[i].used)
			{
				action(m_entries[i].item, context);
			}
		}
	}

private:
	struct Entry
	{
		bool used;
		BASETYPE item;
	};

	uint16_t find(const BASETYPE &item, uint16_t start = 0) const
	{
		uint16_t result = COUNT;

		for (uint16_t i = start; i < m_maxIndex; i++)
		{
			if (m_entries[i].used)
			{
				if (m_entries[i].item == item)
				{
					result = i;
					break;
				}
			}
		}

		return result;
	}

	uint16_t nextEmptySlot(uint16_t start) const
	{
		uint16_t pos = start;

		while (pos < m_maxIndex && m_entries[pos].used)
		{
			pos++;
		}

		return pos;
	}

	uint16_t prevFilledSlot(uint16_t start) const
	{
		uint16_t pos = 0;

		if (m_maxIndex > 0)
		{
			pos = start < m_maxIndex ? start : (m_maxIndex - 1);

			while (pos > 0 && !m_entries[pos].used)
			{
				pos--;
			}
		}

		return pos;
	}

	void insert(uint16_t pos, BASETYPE &item)
	{
		m_entries[pos].item = item;
		m_entries[pos].used = true;
		if (pos + 1 > m_maxIndex)
		{
			m_maxIndex = pos + 1;
		}
	}

	Entry m_entries[COUNT];
	uint16_t m_maxIndex;
};

#endif
