/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef ONESHOTTIMER_H
#define ONESHOTTIMER_H

#include "automaticinterruptlock.h"
#include "eventreceiver.h"
#include "timerservice.h"

/**
 * Countdown timer that sends an event when it has elapsed. A OneShotTimer can be paused using the
 * stop() and start() methods, and reset to the original timeout value using reset().
 * The OneShotTimer needs to be added to a TimerServiceHost instance. The base tick frequency of the
 * TimerServiceHost specifies the rate at which the timeout value is decremented.
 * @param COUNTERTYPE Data type for timeout value, e.g. uint16_t.
 */
template <class COUNTERTYPE>
class OneShotTimer : public TimerService
{
public:
	/**
	 * Constructor.
	 * @param receiver Receiver of the elapsed event.
	 * @param event Event to be sent when timer has elapsed.
	 * @param timeout Timeout value, actual time depends on base tick frequency of used TimerServiceHost.
	 */
	OneShotTimer(EventReceiver *receiver, const Event &event, COUNTERTYPE timeout) :
		m_receiver(receiver), m_event(event), m_startValue(timeout), m_currValue(timeout)
	{
	}

	/**
	 * Gets the currently configured timeout value.
	 * @result Currently configured timeout value.
	 */
	COUNTERTYPE getTimeout(void) const
	{
		AutomaticInterruptLock lock;
		return m_startValue;
	}

	/**
	 * Sets the timeout value that is used when restarting the timer.
	 * @param timeout New timeout value.
	 * @note The changed timeout value takes effect with the next restart only.
	 */
	void setTimeout(COUNTERTYPE timeout)
	{
		AutomaticInterruptLock lock;
		m_startValue = timeout;
	}

	virtual void onTick(void) override
	{
		if (isRunning())
		{
			if (m_currValue > 0)
			{
				if (--m_currValue == 0)
				{
					stop();
					elapsed();
				}
			}
			else
			{
				stop();
			}
		}
	}

	/**
	 * Resets the timeout value to the start value.
	 */
	void reset(void)
	{
		AutomaticInterruptLock lock;
		m_currValue = m_startValue;
	}

	/**
	 * Resets the timeout value to the start value and starts the timer.
	 */
	void restart(void)
	{
		reset();
		start();
	}

protected:
	/**
	 * Method that is called when the timer is running and reaches zero.
	 */
	virtual void elapsed(void)
	{
		if (m_receiver)
		{
			m_receiver->onEvent(m_event);
		}
	}

private:
	EventReceiver * const m_receiver;
	const Event m_event;
	COUNTERTYPE m_startValue;
	COUNTERTYPE m_currValue;
};

#endif
