/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef OUTPUTPORTPIN_H
#define OUTPUTPORTPIN_H

#include <stdint.h>
#include "outputpin.h"
#include "outputport.h"

/**
 * Output pin that can be set to a boolean value.
 * The OutputPortPin works with any OutputPort interface. This could be a simple GpioPort, or a more sophisticated
 * device like an OutputShiftRegisterPort.
 * @note You have to make sure that pins being used for a OutputPortPin to operate on must be configured as output.
 *       The OutputPortPin doesn't know about setting a direction. So if you plan to use a GpioPort for this, first
 *       set it up correctly using its setDirection() method.
 */
class OutputPortPin : public OutputPin
{
public:
	/**
	* Constructor.
	* @param port Pointer to output port the pin is located on.
	* @param index Zero based index of the pin within the port
	*        (e.g. index 2 on PortB results in the pin commonly called PB2).
	*/
	OutputPortPin(OutputPort *port, uint8_t index) :
		m_port(port), m_mask(1u << index)
	{
	}

	virtual void setValue(bool value) override
	{
		m_port->setValue(value ? m_mask : 0, m_mask);
	}

private:
	OutputPort * const m_port;
	const uint8_t m_mask;
};

#endif
