/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef PERIODICTIMER_H
#define PERIODICTIMER_H

#include "oneshottimer.h"

/**
 * Countdown timer that sends an event when it has elapsed and automatically restarts itself.
 * A PeriodicTimer can be paused using the stop() and start() methods, and reset to the original timeout
 * value using reset().
 * The PeriodicTimer needs to be added to a TimerServiceHost instance. The base tick frequency of the
 * TimerServiceHost specifies the rate at which the timeout value is decremented.
 * @param COUNTERTYPE Data type for timeout value, e.g. uint16_t.
 */
template <class COUNTERTYPE>
class PeriodicTimer : public OneShotTimer<COUNTERTYPE>
{
public:
	/**
	 * Constructor.
	 * @param receiver Receiver of the elapsed event.
	 * @param event Event to be sent when timer has elapsed.
	 * @param timeout Timeout value, actual time depends on base tick frequency of used TimerServiceHost.
	 */
	PeriodicTimer(EventReceiver *receiver, const Event &event, COUNTERTYPE timeout) :
		OneShotTimer<COUNTERTYPE>(receiver, event, timeout)
	{
	}

protected:
	virtual void elapsed(void) override
	{
		OneShotTimer<COUNTERTYPE>::elapsed();
		OneShotTimer<COUNTERTYPE>::restart();
	}
};

#endif
