/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef QUEUEDEVENTDISPATCHER_H
#define QUEUEDEVENTDISPATCHER_H

#include <stdint.h>

#include "eventdispatcher.h"
#include "synchronizedringbuffer.h"

/**
 * This component is able to receive events and puts them into a queue.
 * When calling the dispatch() method, the events are sent to all subscribers.
 * Since queuing and dispatching can happen from different contexts, the
 * QueuedEventDispatcher can be used to decouple ISR context and main context.
 * @param SUBSCRIBERCOUNT Maximum number of subscribers.
 * @param QUEUESIZE Maximum number of events that can be queued.
 */
template <uint8_t SUBSCRIBERCOUNT, uint8_t QUEUESIZE>
class QueuedEventDispatcher : public EventDispatcher
{
public:
	/**
	 * Handler for incoming events. The passed event is put into the event queue.
	 * If the queue is full, the event is silently discarded.
	 * @param event Event to handle.
	 */
	virtual void onEvent(const Event &event) override
	{
		m_queue.enqueue(event);
	}

	/**
	 * Dispatches the events from the queue by sending each one to all subscribers.
	 */
	void dispatch(void)
	{
		Event event;

		while (m_queue.dequeue(&event))
		{
			EventDispatcher::onEvent(event);
		}
	}

	/**
	 * Indicates if event queue is empty or requires dispatching.
	 * @result True if event queue is empty, false otherwise.
	 */
	bool isIdle(void)
	{
		return m_queue.isEmpty();
	}

private:
	SynchronizedRingBuffer<Event, QUEUESIZE> m_queue;
};

#endif
