/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef SIMPLEUSART_H
#define SIMPLEUSART_H

#include "usart.h"

/**
 * Simple USART implementation that makes use of hardware units implementing the Usart interface.
 * Read and write class are synchronous. In case non-blocking calls are required, please use
 * BufferedUsart instead.
 * @note The USART device needs to be configured separately, see Usart::Config and Usart::autoFill.
 */
class SimpleUsart
{
public:
	/**
	 * Constructor.
	 * @param usart Pointer to USART device to be used.
	 */
	SimpleUsart(Usart *usart) : m_usart(usart)
	{
	}

	/**
	 * Enables or disables the USART on hardware level.
	 * @param state Desired state of USART, true means enabled, false means disabled.
	 */
	void enable(bool state)
	{
		if (m_usart)
		{
			m_usart->enable(state);
		}
	}

	/**
	 * Read a data byte. This call is blocking and returns only when a data byte was received.
	 * @result Data byte read.
	 */
	uint8_t read(void)
	{
		uint8_t byte = 0;

		if (m_usart)
		{
			while (!m_usart->isReadable())
				;
			byte = m_usart->getData();
		}

		return byte;
	}

	/**
	 * Write a data byte. This call is blocking and returns only when a data byte was placed for sending.
	 * This does not mean that the byte was sent, e.g. when sending another byte the next call will do
	 * the waiting.
	 * @param byte Data byte to send.
	 */
	void write(uint8_t byte)
	{
		if (m_usart)
		{
			while (!m_usart->isWritable())
				;
			m_usart->setData(byte);
		}
	}

protected:
	/**
	 * Pointer to USART device to be used.
	 */
	Usart * const m_usart;
};

#endif
