/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef TIMER_H
#define TIMER_H

#include <stdint.h>

/**
 * Interface for devices that implement an 8 bit timer.
 */
class Timer
{
public:
	/**
	 * Callback type for use with interrupts.
	 */
	class InterruptHandler
	{
	public:
		/**
		 * Handler for Output Compare Match interrupt.
		 */
		virtual void onOutputCompareMatch(void)
		{
		}

		/**
		 * Handler for Overflow interrupt.
		 */
		virtual void onOverflow(void)
		{
		}
	};

	/**
	 * Configuration for timer devices.
	 */
	struct Config
	{
		/**
		 * Operation modes.
		 */
		enum OperationMode
		{
			/// Normal mode (free running timer).
			NORMAL_MODE,
			/// Phase correct PWM.
			PHASE_CORRECT_PWM_MODE,
			/// Clear %Timer on Compare match.
			CTC_MODE,
			/// Fast PWM mode.
			FAST_PWM_MODE,
		};

		/**
		 * Compare output modes.
		 * @note Not all modes are available for each operation mode:
		 * Compare output mode                 | Normal | Phase correct PWM | CTC mode   | Fast PWM
		 * :-----------------------------------|:------:|:-----------------:|:----------:|:-------:
		 * DISCONNECTED                        | Yes    | Yes               | Yes        | Yes
		 * NON_PWM_TOGGLE_ON_COMPARE_MATCH     | Yes    | No                | Yes        | No
		 * NON_PWM_CLEAR_ON_COMPARE_MATCH      | Yes    | No                | Yes        | No
		 * NON_PWM_SET_ON_COMPARE_MATCH        | Yes    | No                | Yes        | No
		 * PHASE_CORRECT_PWM_CLEAR_WHEN_ABOVE  | No     | Yes               | No         | No
		 * PHASE_CORRECT_PWM_SET_WHEN_ABOVE    | No     | Yes               | No         | No
		 * FAST_PWM_NON_INVERTING              | No     | No                | No         | Yes
		 * FAST_PWM_INVERTING                  | No     | No                | No         | Yes
		 */
		enum CompareOutputMode
		{
			/// Normal port operation, OC pin disconnected.
			DISCONNECTED,
			/// Toggle OC pin on compare match.
			NON_PWM_TOGGLE_ON_COMPARE_MATCH,
			/// Clear OC pin on compare match.
			NON_PWM_CLEAR_ON_COMPARE_MATCH,
			/// Set OC pin on compare match.
			NON_PWM_SET_ON_COMPARE_MATCH,
			/// Clear OC pin on compare match when up-counting. Set OC pin on compare match when down-counting.
			PHASE_CORRECT_PWM_CLEAR_WHEN_ABOVE,
			/// Set OC pin on compare match when up-counting. Clear OC pin on compare match when down-counting.
			PHASE_CORRECT_PWM_SET_WHEN_ABOVE,
			/// Clear OC pin on compare match, set OC pin at BOTTOM.
			FAST_PWM_NON_INVERTING,
			/// Set OC pin on compare match, clear OC pin at BOTTOM.
			FAST_PWM_INVERTING,
		};

		/**
		 * Clock sources.
		 */
		enum ClockSource
		{
			/// No clock source (timer stopped).
			NO_CLOCK,
			/// I/O clock without prescaling.
			DIV_1,
			/// I/O clock, prescaler set to 1/8.
			DIV_8,
			/// I/O clock, prescaler set to 1/64.
			DIV_64,
			/// I/O clock, prescaler set to 1/256.
			DIV_256,
			/// I/O clock, prescaler set to 1/1024.
			DIV_1024,
			/// External clock source, clock on falling edge.
			EXT_FALLING_EDGE,
			/// External clock source, clock on rising edge.
			EXT_RISING_EDGE,
		};

		/**
		 * Operation mode of timer.
		 */
		OperationMode operationMode;
		/**
		 * Compare output mode to be used.
		 */
		CompareOutputMode compareOutputMode;
		/**
		 * Clock source to be used.
		 */
		ClockSource clockSource;
		/**
		 * Initial value for output compare register.
		 * Can be modified at runtime using setOutpuCompareValue().
		 */
		uint8_t outputCompareValue;
	};

	/**
	 * Configures the timer device.
	 * @param config Configuration for timer device.
	 */
	virtual void configure(const Config &config) = 0;

	/**
	 * Gets the current value of the counter register.
	 * @result Current value of the counter register.
	 */
	virtual uint8_t getCounterRegister(void) = 0;

	/**
	 * Sets the counter register value.
	 * @param value Value to set.
	 */
	virtual void setCounterRegister(uint8_t value) = 0;

	/**
	 * Sets the output compare register value.
	 * @param value Value to set.
	 */
	virtual void setOutputCompareValue(uint8_t value) = 0;

	/**
	 * Strobes the force output compare match flag.
	 */
	virtual void forceOutputCompareMatch(void) = 0;

	/**
	 * Sets callback for interrupts.
	 * @param handler Pointer to class implementing the interrupt handler interface.
	 */
	virtual void setInterruptHandler(InterruptHandler *handler) = 0;

	/**
	 * Enables or disabled the Output Compare Match interrupt.
	 * @param state True to enable interrupt, false to disable.
	 */
	virtual void enableOutputCompareMatchInterrupt(bool state) = 0;

	/**
	 * Enables or disabled the Overflow interrupt.
	 * @param state True to enable interrupt, false to disable.
	 */
	virtual void enableOverflowInterrupt(bool state) = 0;

	/**
	 * Automatically fills a configuration structure for CTC mode.
	 * @param config Configuration structure to fill.
	 * @param desiredCycleTime Cycle time in microseconds to achieve.
	 * @param actualCycleTime Optional pointer to buffer to receive the cycle time
	 *        that was actually configured.
	 * @result True if configuration structure was filled successfully, false otherwise.
	 * @note Template parameter CPUFREQUENCY must contain the CPU frequency in Hertz.
	 * @note Using compile-time constants for cycleTime reduces code size dramatically.
	 */
	template <uint32_t CPUFREQUENCY>
	static bool autoFill(Config &config, uint32_t desiredCycleTime, uint32_t *actualCycleTime = nullptr)
	{
		bool success = false;

		config.operationMode = Config::CTC_MODE;
		config.compareOutputMode = Config::DISCONNECTED;
		const double desiredDiv = (CPUFREQUENCY / 1000000.0) * desiredCycleTime;
		double actualDiv;

		if (desiredDiv <= 1 * 256ul) // can be created using DIV_1?
		{
			config.clockSource = Config::DIV_1;
			config.outputCompareValue = desiredDiv + 0.5;
			actualDiv = config.outputCompareValue * 1.0;
			success = true;
		}
		else if (desiredDiv <= 8 * 256ul) // can be created using DIV_8?
		{
			config.clockSource = Config::DIV_8;
			config.outputCompareValue = (desiredDiv / 8) + 0.5;
			actualDiv = config.outputCompareValue * 8.0;
			success = true;
		}
		else if (desiredDiv <= 64 * 256ul) // can be created using DIV_64?
		{
			config.clockSource = Config::DIV_64;
			config.outputCompareValue = (desiredDiv / 64) + 0.5;
			actualDiv = config.outputCompareValue * 64.0;
			success = true;
		}
		else if (desiredDiv <= 256 * 256ul) // can be created using DIV_256?
		{
			config.clockSource = Config::DIV_256;
			config.outputCompareValue = (desiredDiv / 256) + 0.5;
			actualDiv = config.outputCompareValue * 256;
			success = true;
		}
		else if (desiredDiv <= 1024 * 256ul) // can be created using DIV_1024?
		{
			config.clockSource = Config::DIV_1024;
			config.outputCompareValue = (desiredDiv / 1024) + 0.5;
			actualDiv = config.outputCompareValue * 1024.0;
			success = true;
		}

		if (actualCycleTime != nullptr)
		{
			*actualCycleTime = actualDiv / (CPUFREQUENCY / 1000000.0);
		}

		return success;
	}
};

#endif
