/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef TIMERSERVICEHOST_H
#define TIMERSERVICEHOST_H

#include <stdint.h>

#include "timer.h"
#include "timerservice.h"
#include "automaticinterruptlock.h"
#include "itemcollection.h"

/**
 * The TimerServiceHost template class is designed to host services derived from Service which can be
 * added and removed at runtime. The host registers as interrupt handler at the Timer instance passed
 * to the constructor, forwarding all Output Compare Match interrupts to the Service::onTick methods.
 * The timer services can implement specific behavior, such as cyclic event creation or time measuring.
 * @param SERVICECOUNT Maximum number of services that can be registered.
 */
template <uint8_t SERVICECOUNT>
class TimerServiceHost : public Timer::InterruptHandler
{
public:
	/**
	 * Constructor.
	 * @param timer Pointer to timer instance used by this TimeServiceHost instance.
	 */
	TimerServiceHost(Timer *timer)
	{
		timer->setInterruptHandler(this);
		timer->enableOutputCompareMatchInterrupt(true);
	}

	/**
	 * Adds a service. A service can be added only once.
	 * @param service Service to add.
	 * @result True if adding was successful, false otherwise.
	 */
	bool add(TimerService *service)
	{
		AutomaticInterruptLock lock;
		return m_services.addUnique(service);
	}

	/**
	 * Removes a service.
	 * @param service Service to remove.
	 * */
	void remove(TimerService *service)
	{
		AutomaticInterruptLock lock;
		m_services.remove(service);
	}

	virtual void onOutputCompareMatch(void) override
	{
		m_services.forEach([](TimerService *&s, const void *ctx){s->onTick();});
	}

private:
	ItemCollection<TimerService*, SERVICECOUNT> m_services;
};

#endif
