/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef INTEGRATEDCIRCUIT_H
#define INTEGRATEDCIRCUIT_H

/**
 * The IntegratedCircuit class is meant as a base class for all sorts of integrated circuits. IntegratedCircuit Itself
 * doesn't contain any functionality, but contains interfaces that can be implemented by derived classes to show that
 * a certain feature is present, e.g. isClocked for circuits that are clocked. The idea is that all circuits that have
 * such a feature can be treated equally and that the names of such methods are not only the same by pure coincidence.
 */
class IntegratedCircuit
{
public:
	/**
	 * Interface for circuits that have a clock input.
	 */
	class isClocked
	{
	public:
		/**
		 * Toggle clock input.
		 */
		virtual void clock(void) = 0;
	};

	/**
	 * Interface for circuits that have an output enable function.
	 */
	class hasOutputEnable
	{
	public:
		/**
		 * Enable or disable outputs.
		 * @param enable Controls if outputs are enabled (true) or disabled (false).
		 * @note The implementing class maps this to the correct logic (active high or active low), e.g.
		 *       enabling outputs by passing true will work in both cases as expected.
		 */
		virtual void enableOutput(bool enable) = 0;
	};

	/**
	 * Interface for circuits that have a clear function.
	 */
	class hasClear
	{
	public:
		/**
		 * Toggle clear input.
		 */
		virtual void clear(void) = 0;
	};

	/**
	 * Interface for circuits that have an output latch function.
	 */
	class hasOutputLatch
	{
	public:
		/**
		 * Toggle clock input for output latch.
		 */
		virtual void latchOutput(void) = 0;
	};
};

#endif
