/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef OUTPUTSHIFTREGISTERPORT_H
#define OUTPUTSHIFTREGISTERPORT_H

#include <stdint.h>

#include "outputport.h"
#include "outputshiftregister.h"

/**
 * Class that implements an OutputPort using a shift register. It uses the OutputShiftRegister interface that defines
 * only a data and clock line, thus there will be always 8 bits transported (instead of clearing the register and
 * making use of leading zeros).
 * You can define OutputPortPin instances on top of the OutputPort which will work just like regular OutputPins, just
 * a bit slower. When having the property autoFlush set, each write access to the OutputShiftRegisterPort will
 * directly be transferred to the associated device. If autoFlush is not set, multiple write actions can be
 * accumulated before transferring them to the device explicitly using the flush() method.
 */
class OutputShiftRegisterPort : public OutputPort
{
public:
	/**
	 * Constructor.
	 * @param shiftRegister Pointer to OutputShiftRegister instance to be used as output device.
	 * @param autoFlush Value of autoFlush property, defaulting to true.
	 */
	OutputShiftRegisterPort(OutputShiftRegister *shiftRegister, bool autoFlush = true) :
		m_shiftRegister(shiftRegister), m_autoFlush(autoFlush), m_value(0)
	{
	}

	/**
	 * Getter for current value of autoFlush property.
	 * @result Current value of autoFlush property.
	 */
	bool getAutoFlush(void) const
	{
		return m_autoFlush;
	}

	/**
	 * Setter for value of autoFlush property.
	 * @param state New value for autoFlush property, true means that autoFlush is active.
	 */
	void setAutoFlush(bool state)
	{
		m_autoFlush = state;
	}

	/**
	 * Setter for 8 bit OutputPort value. Depending on the state of the autoFlush property this directly triggers
	 * a writing operation to the associated device (autoFlush = true), or updates the internal value only
	 * (autoFlush = false). When autoFlush is inactive, the value must be transferred explicitly using flush().
	 * @param value Value to be set.
	 * @param mask Mask with pins to manipulate according to value.
	 */
	virtual void setValue(uint8_t value, uint8_t mask = 0xff) override
	{
		m_value = (m_value & ~mask) | (value & mask);

		if (m_autoFlush)
		{
			flush();
		}
	}

	/**
	 * Excplicitly transferrs internal value to the associated device.
	 */
	void flush(void)
	{
		uint8_t shiftMask = 0x80;
		while (shiftMask)
		{
			m_shiftRegister->setData(m_value & shiftMask);
			shiftMask = shiftMask >> 1;
			m_shiftRegister->clock();
		}
	}

protected:
	/**
	 * Pointer to OutputShiftRegister instance to be used as output device.
	 */
	OutputShiftRegister * const m_shiftRegister;

private:
	bool m_autoFlush;
	uint8_t m_value;
};

#endif
