/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "automaticinterruptlock.h"
#include "int0.h"

static volatile ExternalInterrupt::InterruptHandler *Int0_InterruptHandler;

ISR(INT0_vect)
{
	if (Int0_InterruptHandler)
	{
		const_cast<ExternalInterrupt::InterruptHandler*>(Int0_InterruptHandler)->onExternalInterrupt();
	}
}

Int0::Int0(void)
{
	// Safe defaults: low level trigger, no interrupts
	MCUCR &= ~((1 << ISC01) | (1 << ISC00));
	GICR &= ~(1 << INT0);
	Int0_InterruptHandler = nullptr;
}

void Int0::setTriggerMode(TriggerMode mode)
{
	const uint8_t keepMask = ~((1 << ISC01) | (1 << ISC00));
	uint8_t setMask;

	switch (mode)
	{
	default:
	case LOW_LEVEL:
		setMask = (0 << ISC01) | (0 << ISC00);
		break;
	case ANY_CHANGE:
		setMask = (0 << ISC01) | (1 << ISC00);
		break;
	case FALLING_EDGE:
		setMask = (1 << ISC01) | (0 << ISC00);
		break;
	case RISING_EDGE:
		setMask = (1 << ISC01) | (1 << ISC00);
		break;
	}
	MCUCR = (MCUCR & keepMask) | setMask;
}

void Int0::enableInterrupt(bool state)
{
	if (state)
	{
		GICR |= (1 << INT0);
	}
	else
	{
		GICR &= ~(1 << INT0);
	}
}

void Int0::setInterruptHandler(InterruptHandler *handler)
{
	AutomaticInterruptLock lock;
	Int0_InterruptHandler = handler;
}
