/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "automaticinterruptlock.h"
#include "timer0.h"

static volatile Timer::InterruptHandler *Timer0_InterruptHandler;

ISR(TIMER0_COMP_vect)
{
	if (Timer0_InterruptHandler)
	{
		const_cast<Timer::InterruptHandler*>(Timer0_InterruptHandler)->onOutputCompareMatch();
	}
}

ISR(TIMER0_OVF_vect)
{
	if (Timer0_InterruptHandler)
	{
		const_cast<Timer::InterruptHandler*>(Timer0_InterruptHandler)->onOverflow();
	}
}

Timer0::Timer0(void)
{
	TCCR0 = 0;
	TIMSK &= ~((1 << OCIE0) | (1 << TOIE0));
	Timer0_InterruptHandler = nullptr;
}

void Timer0::configure(const Config &config)
{
	setOperationMode(config.operationMode);
	setCompareOutputMode(config.compareOutputMode);
	setClockSource(config.clockSource);
	setOutputCompareValue(config.outputCompareValue);
}

uint8_t Timer0::getCounterRegister(void)
{
	return TCNT0;
}

void Timer0::setCounterRegister(uint8_t value)
{
	TCNT0 = value;
}

void Timer0::setOutputCompareValue(uint8_t value)
{
	OCR0 = value;
}

void Timer0::forceOutputCompareMatch(void)
{
	TCCR0 |= (1 << FOC0);
}

void Timer0::setInterruptHandler(InterruptHandler *handler)
{
	AutomaticInterruptLock lock;
	Timer0_InterruptHandler = handler;
}

void Timer0::enableOutputCompareMatchInterrupt(bool state)
{
	if (state)
	{
		TIMSK |= (1 << OCIE0);
	}
	else
	{
		TIMSK &= ~(1 << OCIE0);
	}
}

void Timer0::enableOverflowInterrupt(bool state)
{
	if (state)
	{
		TIMSK |= (1 << TOIE0);
	}
	else
	{
		TIMSK &= ~(1 << TOIE0);
	}
}

void Timer0::setOperationMode(Config::OperationMode mode)
{
	const uint8_t keepMask = ~((1 << WGM01) | (1 << WGM00));
	uint8_t setMask;

	switch (mode)
	{
	default:
	case Config::NORMAL_MODE:
		setMask = (0 << WGM01) | (0 << WGM00);;
		break;
	case Config::PHASE_CORRECT_PWM_MODE:
		setMask = (0 << WGM01) | (1 << WGM00);
		break;
	case Config::CTC_MODE:
		setMask = (1 << WGM01) | (0 << WGM00);
		break;
	case Config::FAST_PWM_MODE:
		setMask = (1 << WGM01) | (1 << WGM00);
		break;
	}
	TCCR0 = (TCCR0 & keepMask) | setMask;
}

void Timer0::setClockSource(Config::ClockSource source)
{
	const uint8_t keepMask = ~((1 << CS02) | (1 << CS01) | (1 << CS00));
	uint8_t setMask;

	switch (source)
	{
	default:
	case Config::NO_CLOCK:
		setMask = (0 << CS02) | (0 << CS01) | (0 << CS00);
		break;
	case Config::DIV_1:
		setMask = (0 << CS02) | (0 << CS01) | (1 << CS00);
		break;
	case Config::DIV_8:
		setMask = (0 << CS02) | (1 << CS01) | (0 << CS00);
		break;
	case Config::DIV_64:
		setMask = (0 << CS02) | (1 << CS01) | (1 << CS00);
		break;
	case Config::DIV_256:
		setMask = (1 << CS02) | (0 << CS01) | (0 << CS00);
		break;
	case Config::DIV_1024:
		setMask = (1 << CS02) | (0 << CS01) | (1 << CS00);
		break;
	case Config::EXT_FALLING_EDGE:
		setMask = (1 << CS02) | (1 << CS01) | (0 << CS00);
		break;
	case Config::EXT_RISING_EDGE:
		setMask = (1 << CS02) | (1 << CS01) | (1 << CS00);
		break;
	}
	TCCR0 = (TCCR0 & keepMask) | setMask;
}

void Timer0::setCompareOutputMode(Config::CompareOutputMode mode)
{
	const uint8_t keepMask = ~((1 << COM01) | (1 << COM00));
	uint8_t setMask;

	switch (mode)
	{
	default:
	case Config::DISCONNECTED:
		setMask = (0 << COM01) | (0 << COM00);
		break;
	case Config::NON_PWM_TOGGLE_ON_COMPARE_MATCH:
		setMask = (0 << COM01) | (1 << COM00);
		break;
	case Config::NON_PWM_CLEAR_ON_COMPARE_MATCH:
		setMask = (1 << COM01) | (0 << COM00);
		break;
	case Config::NON_PWM_SET_ON_COMPARE_MATCH:
		setMask = (1 << COM01) | (1 << COM00);
		break;
	case Config::PHASE_CORRECT_PWM_CLEAR_WHEN_ABOVE:
		setMask = (1 << COM01) | (0 << COM00);
		break;
	case Config::PHASE_CORRECT_PWM_SET_WHEN_ABOVE:
		setMask = (1 << COM01) | (1 << COM00);
		break;
	case Config::FAST_PWM_NON_INVERTING:
		setMask = (1 << COM01) | (0 << COM00);
		break;
	case Config::FAST_PWM_INVERTING:
		setMask = (1 << COM01) | (1 << COM00);
		break;
	}
	TCCR0 = (TCCR0 & keepMask) | setMask;
}
