/*
 * Copyright (c) 2019, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "automaticinterruptlock.h"
#include "usart0.h"

static volatile Usart::InterruptHandler *Usart0_InterruptHandler;

ISR(USART_UDRE_vect)
{
	if (Usart0_InterruptHandler)
	{
		const_cast<Usart::InterruptHandler*>(Usart0_InterruptHandler)->onDataRegisterEmpty();
	}
}

ISR(USART_RXC_vect)
{
	if (Usart0_InterruptHandler)
	{
		const_cast<Usart::InterruptHandler*>(Usart0_InterruptHandler)->onRxComplete();
	}
}

Usart0::Usart0(void)
{
	UCSRB = 0;
	Usart0_InterruptHandler = nullptr;
}

void Usart0::enable(bool state)
{
	if (state)
	{
		UCSRB |= (1 << RXEN) | (1 << TXEN);
	}
	else
	{
		UCSRB &= ~((1 << RXEN) | (1 << TXEN));
	}
}

void Usart0::configure(const Config &config)
{
	setBaudRateSetting(config.baudRateSetting);
	setParity(config.parity);
	setStopBits(config.stopBits);
	setDataBits(config.dataBits);
}

void Usart0::setData(uint8_t data)
{
	UDR = data;
}

uint8_t Usart0::getData(void)
{
	return UDR;
}

bool Usart0::isReadable(void)
{
	return UCSRA & (1 << RXC);
}

bool Usart0::isWritable(void)
{
	return UCSRA & (1 << UDRE);
}

void Usart0::setInterruptHandler(Usart::InterruptHandler *handler)
{
	AutomaticInterruptLock lock;
	Usart0_InterruptHandler = handler;
}

void Usart0::enableDataRegisterEmptyInterrupt(bool state)
{
	if (state)
	{
		UCSRB |= (1 << UDRIE);
	}
	else
	{
		UCSRB &= ~(1 << UDRIE);
	}
}

void Usart0::enableRxCompleteInterrupt(bool state)
{
	if (state)
	{
		UCSRB |= (1 << RXCIE);
	}
	else
	{
		UCSRB &= ~(1 << RXCIE);
	}
}

// Special procedure to read UCSRC. Lost half an hour here.
static uint8_t getUCSRC(void)
{
	AutomaticInterruptLock lock;

	uint8_t ucsrc = UBRRH;
	ucsrc = UCSRC;
	return ucsrc;
}

void Usart0::setBaudRateSetting(const Config::BaudRateSetting &baudRateSetting)
{
	UBRRH = baudRateSetting.baudRateRegister >> 8;
	UBRRL = baudRateSetting.baudRateRegister;
	if (baudRateSetting.doubleSpeed)
	{
		UCSRA |= (1 << U2X);
	}
	else
	{
		UCSRA &= ~(1 << U2X);
	}
}

void Usart0::setDataBits(Config::DataBits dataBits)
{
	uint8_t ucsrc = getUCSRC() & ~(0x03 << UCSZ0);

	switch (dataBits)
	{
	case Config::DATABITS_5:
		UCSRC = ucsrc;
		break;
	case Config::DATABITS_6:
		UCSRC = ucsrc | (1 << UCSZ0);
		break;
	case Config::DATABITS_7:
		UCSRC = ucsrc | (1 << UCSZ1);
		break;
	case Config::DATABITS_8:
		UCSRC = ucsrc | (1 << UCSZ1) | (1 << UCSZ0);
		break;
	}
}

void Usart0::setParity(Config::Parity parity)
{
	uint8_t ucsrc = getUCSRC() & ~(3 << UPM0);

	switch (parity)
	{
	case Config::PARITY_DISABLED:
		UCSRC = ucsrc;
		break;
	case Config::PARITY_ODD:
		UCSRC = ucsrc | (1 << UPM1);
		break;
	case Config::PARITY_EVEN:
		UCSRC = ucsrc | (1 << UPM1) | (1 << UPM0);
		break;
	}
}

void Usart0::setStopBits(Config::StopBits stopBits)
{
	uint8_t ucsrc = getUCSRC() & ~(1 << USBS);

	switch (stopBits)
	{
	case Config::ONE_STOPBIT:
		UCSRC = ucsrc;
		break;
	case Config::TWO_STOPBITS:
		UCSRC = ucsrc | (1 << USBS);
		break;
	}
}
